#! /usr/bin/env python

# Produces a .api file for SciTE's identifier completion and calltip features.
# invoke as 
# 	python tags2api.py tags > x.api
# before running this program, create a tags file with
# 	ctags --excmd=number --c-types=pcdgstu <header files>

import string
import time
import sys
import getopt

# Definitions that start with _ are often used for infrastructure
# like include guards and can be removed with
removePrivate=1

class FileCache:
    '''Caches the contents of a set of files.
    Avoids reading files repeatedly from disk by holding onto the
    contents of each file as a list of strings.
    '''

    def __init__(self):
        self.filecache = {}
        
    def grabFile(self, filename):
        '''Return the contents of a file as a list of strings.
        New line characters are removed.
        '''
        if not self.filecache.has_key(filename):
            contents=[]
            f = open(filename)
            for line in f.readlines():
                if line[-1:] == '\n': line = line[:-1]
                contents.append(line)
            f.close()
            self.filecache[filename] = contents
        return self.filecache[filename]

def bracesDiff(s):
    ''' Counts the number of '(' and ')' in a string and returns the difference between the two.
    Used to work out when a function prototype is complete.
    '''
    diff = 0
    mode=0 # 0 <=> default, 1 <=> comment, 2 <=> string
    for i in range(len(s)):
        if mode == 0: # default mode
            if s[i]=='(':
                diff += 1
            elif s[i]==')':
                diff -= 1
            elif s[i]=='"':
                mode=2
            elif i>0 and s[i-1]=='/' and s[i]=='/':
                return diff
            elif i>0 and s[i-1]=='/' and s[i]=='*':
                mode=1
        elif mode == 1: # comment
            if i>0 and s[i-1]=='*' and s[i]=='/':
                mode=0
        elif mode == 2: # string
            if s[i]=='"':
                mode=0
    return diff


def macroDefinition(entityName,curDef):
    if curDef[:8] == '#define ':
        curDef = curDef[8:]
        
        if "(" in curDef and ")" in curDef:
            space = curDef.find(" ")
            startParen = curDef.find(")")
            
            if ( space == -1 ) or ( space > startParen ):
                endParen = curDef.find(")")
                curDef = curDef[:endParen+1:]
                return curDef
    return entityName

def functionDefiniton(entityName,curDef):
    # Normalise the appearance of the prototype.
    curDef = string.strip(curDef)
    # Replace whitespace sequences with a single space character.
    curDef = string.join(string.split(curDef))
    # Remove space around the '('.
    curDef = string.replace(string.replace(curDef, " (", '('), "( ", '(')
    # Remove trailing semicolon.
    curDef = string.replace(curDef, ";", '')
    # Remove implementation if present.
    if "{" in curDef and "}" in curDef:
        startImpl = curDef.find("{")
        endImpl = curDef.find("}")
        curDef = curDef[:startImpl] + curDef[endImpl+1:]
    else:
        # Remove trailing brace.
        curDef = curDef.rstrip("{")
    # Remove return type.
    curDef = curDef[string.find(curDef, entityName):]
    # Remove virtual indicator.
    if curDef.replace(" ", "").endswith(")=0"):
        curDef = curDef.rstrip("0 ")
        curDef = curDef.rstrip("= ")
    # Remove trailing space.
    curDef = curDef.rstrip()

    return curDef

def convert ( filenames, format ):
    fc = FileCache()
    apis = set()
    
    for filename in filenames:
        infile = open(filename, 'r')
        
        for line in infile:
            if line[0] != '!':	# Not a comment.
                (entityName, fileName, lineNo, tagType) = string.split(line, "\t")[:4]
                
                try:
                    curLineNo = string.atoi(lineNo[:-2]) - 1	
                    # -1 because line numbers in tags file start at 1.
                except:
                    continue
                contents = fc.grabFile(fileName)
                
                if (not removePrivate or entityName[0] != '_') and not entityName.startswith("operator "):
                    if tagType[0] in "pf":	# Function prototype.
                        try:
                            braces = bracesDiff(contents[curLineNo])
                            curDef = contents[curLineNo]
                            
                            # Search for end of prototype.
                            while braces > 0:	
                                curLineNo = curLineNo + 1
                                braces = braces + bracesDiff(contents[curLineNo])
                                curDef = curDef + contents[curLineNo]

                            apis.add(format(functionDefiniton(entityName,curDef),curLineNo,fileName))
                        except IndexError:
                            pass
                    elif tagType[0] == 'd':	# Macro definition.
                        curDef = contents[curLineNo]
                        apis.add(format(macroDefinition(entityName,curDef),curLineNo,fileName))
                        
                    elif tagType[0] == 't':	# Type definition.
                        curDef = contents[curLineNo]
                        apis.add(format(entityName,curLineNo,fileName))
                        
                    #~ else:
                        #~ apis.add(format(entityName))
    return apis
    
def plainDefs(curDef,curLineNo,fileName):
    return curDef
    
def withLines(curDef,curLineNo,fileName):
    return "%s:%d: %s" % (fileName,curLineNo+1,curDef)

def main ():
    # parse command line options
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hl", ["help","lines"])
    except getopt.error, msg:
        print msg
        print "for help use --help"
        sys.exit(2)
    
    format = plainDefs
    
    # process options
    for o, a in opts:
        if o in ("-h", "--help"):
            print __doc__
            sys.exit(0)
        if o in ("-l", "--lines"):
            format = withLines
        
    apis = convert(args,format)
    print "\n".join(sorted(list(apis)))

if __name__ == "__main__":
    main()